#include"zega.h"
#include<math.h>
#include<qpe/qcopenvelope_qws.h>

QPoint rotate(QPoint point, double deg) {
	QPoint rtnPoint;
	double rad = deg * M_PI / 180.0;
	rtnPoint.setX(point.x() * cos(rad) * -1 + point.y() * sin(rad));
	rtnPoint.setY((point.x() * sin(rad) + point.y() * cos(rad)) * -1);
	return rtnPoint;
}

Config Zega::config = Config("Zega");

Zega::Zega(QWidget* parent, const char* name) :
	QWidget(parent, name) {
	// ɤ߹
	config.setGroup("Color");
	_routeColor.setNamedColor("#FF0000");
	if (!config.hasKey("routeColor"))
		config.writeEntry("routeColor", "#FF0000");
	else
		_routeColor.setNamedColor(config.readEntry("routeColor"));

	_recentLogColor.setNamedColor("#FFFF00");
	if (!config.hasKey("recentLogColor"))
		config.writeEntry("recentLogColor", "#FFFF00");
	else
		_recentLogColor.setNamedColor(config.readEntry("recentLogColor"));

	_obsoluteLogColor.setNamedColor("#00FFFF");
	if (!config.hasKey("obsoluteLogColor"))
		config.writeEntry("obsoluteLogColor", "#00FFFF");
	else
		_obsoluteLogColor.setNamedColor(config.readEntry("obsoluteLogColor"));

	_currentPointColor.setNamedColor("#00FF00");
	if (!config.hasKey("currentPointColor"))
		config.writeEntry("currentPointColor", "#00FF00");
	else
		_currentPointColor.setNamedColor(config.readEntry("currentPointColor"));

	// 
	config.setGroup("Width");
	_routeWidth = 6;
	if (!config.hasKey("routeWidth"))
		config.writeEntry("routeWidth", 6);
	else
		_routeWidth = config.readNumEntry("routeWidth");
	if (_routeWidth < 1)
		_routeWidth = 1;

	_logWidth = 6;
	if (!config.hasKey("logWidth"))
		config.writeEntry("logWidth", 6);
	else
		_logWidth = config.readNumEntry("logWidth");
	if (_logWidth < 1)
		_logWidth = 1;

	// ɽե饰
	_autoCentered = true;
	_showGpsLog = true;
	_showRoute = true;

	// ȥ
	setCaption(tr("Zega"));

	// ˥塼Сɲ
	_menu = new QMenuBar(this, "menubar");

	_filemenu = new QPopupMenu(this, "filemenu");
	_filemenu->insertItem(tr("&Load File"), this, SLOT(slotLoad()));
	_filemenu->insertSeparator();
	_filemenu->insertItem(tr("&Quit"), _app, SLOT(quit()));

	// Ͽ޾̵Ͻ
	config.setGroup("Maps");
	// ХޥåNavi̵Ͻ
	if ( !config.hasKey("useMobileMapNaviMap") )
		config.writeEntry("useMobileMapNaviMap", "true");

	_mobileMapNaviMaps = NULL;
	if (config.readEntry("useMobileMapNaviMap") == "true") {
		if ( !config.hasKey("mobileMapNaviMapDir") ) {
			config.writeEntry("mobileMapNaviMapDir", "/mnt/card/sd_map");

			if ( !config.hasKey("mapNum") )
				config.writeEntry("mapNum", 0);
			config.writeEntry("mapNum", 8);
			config.writeEntry("map0", tr("ХޥåNavi,M,0"));
			config.writeEntry("map1", tr("ХޥåNaviϿ,M,1"));
			config.writeEntry("map2", tr("ХޥåNaviϿ,M,2"));
			config.writeEntry("map3", tr("ХޥåNaviϿ,M,3"));
			config.writeEntry("map4", tr("ХޥåNaviܺϿ,M,4"));
			config.writeEntry("map5", tr("1/150000,C,/mnt/card/maps/1"));
			config.writeEntry("map6", tr("1/21000,C,/mnt/card/maps/2"));
			config.writeEntry("map7", tr("1/8000,C,/mnt/card/maps/3"));
		}
		QDir mmnDir(config.readEntry("mobileMapNaviMapDir"));
		if (mmnDir.exists()) {
			_mobileMapNaviMaps = new MobileMapNaviMaps(config.readEntry("mobileMapNaviMapDir"));
		} else {
			config.writeEntry("useMobileMapNaviMap", "false");
		}
	}

	// Ͽ˥塼
	_scalemenu = new QPopupMenu(this, "scalemenu");
	int mapNum = config.readNumEntry("mapNum");
	QString strTmp;
	QString strNumTmp;
	int intMapId = 0;
	mapDirs.clear();
	for (int i = 0; i < mapNum; i++) {
		strTmp = QString("map") + strNumTmp.setNum(i);
		if (config.hasKey(strTmp) ) {
			QStringList mapInfo = config.readListEntry(strTmp, ',');
			if (mapInfo.count() >= 3) {
				if (_mobileMapNaviMaps != NULL || mapInfo[1] != "M") {
					mapDirs.append(config.readEntry(strTmp));
					int itemId = _scalemenu->insertItem(mapInfo[0], this, SLOT(slotChScale(int)));
					_scalemenu->setItemParameter(itemId, intMapId);
					intMapId++;
				}
			}
		}
	}

	_gpsmenu = new QPopupMenu(this, "gpsmenu");
	_gpsmenu->insertItem(tr("&Start GPS"), this, SLOT(startGps()));
	_gpsmenu->insertItem(tr("S&top GPS"), this, SLOT(stopGps()));

	_displaymenu = new QPopupMenu(this, "displaymenu");
	_displaymenu->insertItem(tr("&GPS Log On/Off"), this,
			SLOT(slotToggleShowGpsLog()));
	_displaymenu->insertItem(tr("&Route On/Off"), this,
			SLOT(slotToggleShowRoute()));
	_displaymenu->insertItem(tr("&Auto Centered On/Off"), this,
			SLOT(slotToggleAutoCentered()));

	_helpmenu = new QPopupMenu(this, "helpmenu");
	_helpmenu->insertItem(tr("&About"), this, SLOT(about()));

	_menu->insertItem(tr("&File"), _filemenu);
	_menu->insertItem(tr("&Scale"), _scalemenu);
	_menu->insertItem(tr("&GPS"), _gpsmenu);
	_menu->insertItem(tr("&Display"), _displaymenu);
	_menu->insertItem(tr("&Help"), _helpmenu);

	config.setGroup("CurrentCoordinate");
	// ֤߰¸ߤʤؤ
	if (!config.hasKey("lat"))
		config.writeEntry("lat", "N35.40.40.00");
	if (!config.hasKey("lon"))
		config.writeEntry("lon", "E139.46.15.00");

	_centerNS.parse(config.readEntry("lat"));
	_centerEW.parse(config.readEntry("lon"));
	_pixelNS = 0;
	_pixelEW = 0;
	_pixelBase = PIXEL_BASE;

	// Ͽ
	_mapIndex = 0;
	config.setGroup("Maps");
	if (!config.hasKey("currentMap"))
		config.writeEntry("currentMap", 0);
	else {
		_mapIndex = config.readNumEntry("currentMap");
		if (_mapIndex >= mapDirs.count())
			_mapIndex = 0;
	}

	// ޥåץå
	if (!config.hasKey("mapCacheSize"))
		config.writeEntry("mapCacheSize", 12);
	int mapCacheSize = config.readNumEntry("mapCacheSize");
	_mapCache = new MapCache(mapCacheSize);

	_mapList = &_customMapList;
	_buffer = NULL;
	_mapBuffer = NULL;
	slotChScale(_mapIndex);

	// GPS
	config.setGroup("GPS");
	// GPSեѥ
	if ( !config.hasKey("logDir") )
		config.writeEntry("logDir", "/mnt/card/gpslog");
	// GPSǥХե
	if ( !config.hasKey("devFile") )
		config.writeEntry("devFile", "/dev/ttyS3");
	// NMEA-0183Ϥ¬Ϸ
	if ( !config.hasKey("gpsOutput") )
		config.writeEntry("gpsOutput", "WGS84");

	// ľȤGPS־򵭲ֳ
	if ( !config.hasKey("recentTerm") )
		config.writeEntry("recentTerm", "1");
	int recentTerm = config.readNumEntry("recentTerm");
	if (recentTerm <= 0)
		recentTerm = 1;

	// ľȤGPS־Ͽ
	if ( !config.hasKey("recentLogNum") )
		config.writeEntry("recentLogNum", "600");
	int recentLogNum = config.readNumEntry("recentLogNum");
	if (recentLogNum <= 0)
		recentLogNum = 600;

	// ŤȤGPS־򵭲ֳ
	if ( !config.hasKey("obsolateTerm") )
		config.writeEntry("obsolateTerm", "10");
	int obsolateTerm = config.readNumEntry("obsolateTerm");
	if (obsolateTerm <= recentTerm)
		obsolateTerm = recentTerm * 4;

	_gpsLogHandler = new GpsLogHandler(recentTerm, recentLogNum, obsolateTerm);
	_gpsLogHandler->setLogDir(config.readEntry("logDir"));
	_gpsLogHandler->setDevFile(config.readEntry("devFile"));
	_gpsLogHandler->setGpsOutput(config.readEntry("gpsOutput"));

	// ưǥ󥿥󥰤볰Ȥ
	if ( !config.hasKey("autoCenteredBounds") )
		config.writeEntry("autoCenteredBounds", 50);
	_autoCenteredBounds = config.readNumEntry("autoCenteredBounds");
	if (_autoCenteredBounds <= 0)
		_autoCenteredBounds = 0;

	_mapBuffer = new QPixmap();

	// ̹ޡ
	_updateTimer = new QTimer( this );
	connect(_updateTimer, SIGNAL(timeout()), this, SLOT(slotRePaint()) );
}

Zega::~Zega() {
	if (_gpsLogHandler->isAlive())
		stopGps();

	config.setGroup("CurrentCoordinate");
	config.writeEntry("lat", _centerNS.toNS());
	config.writeEntry("lon", _centerEW.toEW());
	config.setGroup("Maps");
	config.writeEntry("currentMap", (int)_mapIndex);
}

void Zega::disableSuspend() {
	config.setGroup("General");
	if ( !config.hasKey("controlZaurusSuspend") )
		config.writeEntry("controlZaurusSuspend", "true");

	if (config.readEntry("controlZaurusSuspend") != "true")
		return;

	// ưȼưڥɤ̵ˤ
	/* +--+--+ڥ+
	 * |0 |      |          |
	 * |1 |      |    o     |
	 * |2 |  o   |          |
	 * |3 |  o   |    o     |
	 */
	QCopEnvelope("QPE/System", "setScreenSaverMode(int)") << 0;
	// ꡼󥻡Ф̵ˤ
	QCopEnvelope("QPE/System", "setScreenSaverIntervals(int,int,int)") << 0
			<< 0 << 0;
}

void Zega::enableSuspend() {
	config.setGroup("General");
	if (config.readEntry("controlZaurusSuspend") != "true")
		return;

	// ꡼󥻡Фͭˤ
	QCopEnvelope("QPE/System", "setScreenSaverIntervals(int,int,int)") << -1
			<< -1 << -1;
	// ưȥȥѥդͭˤ
	QCopEnvelope("QPE/System", "setScreenSaverMode(int)") << 3;
}

void Zega::slotChScale(int mapId) {
	_mapIndex = mapId;
	const QStringList& qslTemp = QStringList::split(",", mapDirs[mapId]);
	if (qslTemp[1] == "C") {
		cout << qslTemp[2] << ":" << mapDirs[mapId] << endl;
		_customMapList.changeMap(qslTemp[2]);
		_mapList = &_customMapList;
	} else if (qslTemp[1] == "M") {
		cout << "MobilemapNavi map data" << endl;
		if (_mobileMapNaviMaps == NULL)
			return;
		cout << qslTemp[2] << ":" << mapDirs[mapId] << endl;
		_mobileMapNaviMaps->changeScale(qslTemp[2]);
		_mapList = _mobileMapNaviMaps;
	} else {
		return;
	}

	if (_mapBuffer != NULL)
		paintMaps(_centerNS, _centerEW);
}

void Zega::slotRePaint() {
	static int beforeX = -50;
	static int beforeY = -50;
	Coordinate coordinate;

	if (_gpsLogHandler->isAlive() == false) {
		QPainter painter(this);
		painter.setPen(QPen(black, 1, SolidLine));
		painter.setBrush(QBrush(black, SolidPattern));
		painter.drawEllipse(10, 40, 10, 10);
		painter.end();
		return;
	}

	// ּ߰
	coordinate = _gpsLogHandler->getNowCoordinate();
	double direction = _gpsLogHandler->getNowDirection();
	int x = ((coordinate.getRadEW().toLong() - _centerEW.toLong()) << _pixelBase)/ _pixelEW
			+ this->width()/2;
	int y = ((coordinate.getRadNS().toLong() - _centerNS.toLong()) << _pixelBase)/ _pixelNS
			+ this->height()/2;

	// ֤߰ϰϳξϸ֤߰ѹ
	if (_autoCentered) {
		if ( x < _autoCenteredBounds || x > this->width() -_autoCenteredBounds || y < _autoCenteredBounds || y > this->height() - _autoCenteredBounds) {
			_centerNS = coordinate.getRadNS();
			_centerEW = coordinate.getRadEW();
			paintMaps(_centerNS, _centerEW);
			x = ((coordinate.getRadEW().toLong() - _centerEW.toLong()) << _pixelBase)/ _pixelEW + this->width()/2;
			y = ((coordinate.getRadNS().toLong() - _centerNS.toLong()) << _pixelBase)/ _pixelNS + this->height()/2;
		}
	}

	// 
	if (beforeX >= 0 && beforeX < this->width() && beforeY >= 0 && beforeY
			< this->height()) {
		bitBlt(this, beforeX-27, beforeY-27, _mapBuffer, beforeX-27,
				beforeY-27, 55, 55);
	}

	// ֤߰Ͽޤ˥ݥ
	if (x >= 0 && x < this->width() && y >= 0 && y < this->height()) {
		// ϰϵϿ
		beforeX = x;
		beforeY = y;

		// ֤߰ޤǤεפХåե˵Ͽ
		QPainter painter(_mapBuffer);
		painter.setPen(QPen(blue, 1, SolidLine));
		painter.setBrush(QBrush(_recentLogColor, SolidPattern));
		painter.drawEllipse(x - _logWidth/2, y -_logWidth/2, _logWidth, _logWidth);
		painter.end();

		// 		
		painter.begin(this);
		painter.translate(x, y);
		painter.setPen(QPen(black, 1, SolidLine));
		painter.setBrush(QBrush(_currentPointColor, SolidPattern));
		QPointArray tempPoints(3);
		tempPoints.setPoint(0, rotate(QPoint(0, 10), direction));
		tempPoints.setPoint(1, rotate(QPoint(-5, -9), direction));
		tempPoints.setPoint(2, rotate(QPoint(5, -9), direction));
		painter.drawPolygon(tempPoints);
		painter.setPen(QPen(black, 1, Qt::DashLine));
		painter.setBrush(Qt::NoBrush);
		painter.drawEllipse(-25, -25, 51, 51);		
		painter.end();
	}
	QPainter painter(this);
	// GPS򥢥ƥ֤ɽ
	painter.setPen(QPen(red, 1, SolidLine));
	painter.setBrush(QBrush(red, SolidPattern));
	painter.drawEllipse(10, 40, 10, 10);
	// ־ɽ
	painter.setPen(QPen(black, 1, SolidLine));
	painter.setBrush(QBrush(white, SolidPattern));
	painter.setFont(QFont("courier", 12));
	painter.drawRect(this->width() -150, 40, 140, 44);
	painter.drawText(this->width() -148, 54, coordinate.getRadNS().toNS());
	painter.drawText(this->width() -148, 68, coordinate.getRadEW().toEW());
	painter.drawText(this->width() -148, 82, "Direction: " + QString::number(
			direction, 'f', 2));
	painter.end();
}

void Zega::startGps() {
	config.setGroup("GPS");
	if ( !config.hasKey("UpdateInterval") ) {
		config.writeEntry("UpdateInterval", 1000);
	}
	int intInterval = config.readNumEntry("UpdateInterval");

	QDir dir;
	QString logDir = _gpsLogHandler->getLogDir();
	if (!dir.exists(logDir)) {
		int rtn = QMessageBox::warning(this, tr("GPS Log Dir Not Exist."),
				tr("GPSǥ쥯ȥ(" + logDir + ")ʤ衣\näƤ"), tr(""),
				tr(""), QString::null, 1);
		if (rtn == 0)
		QDir().mkdir(logDir);
	}
	if (!dir.exists(logDir)) {
		QMessageBox::critical(this, tr("Cannot Start GPS"),
		tr("GPSǥ쥯ȥ(" + logDir + ")ʤ\nGPS򤢤ޤ͡"));
		return;
	}

	// LCD꡼󥻡Сzaurusڥ̵
	disableSuspend();

	_gpsLogHandler->start();
	_updateTimer->start(intInterval, FALSE);
}

void Zega::stopGps() {
	_updateTimer->stop();
	_gpsLogHandler->stop();
	cout << "Cardctl end." << endl;

	// zaurusڥᤷ
	enableSuspend();
	paintMaps(_centerNS, _centerEW);
}

void Zega::slotLoad() {
	QTKFileDialog dialog(this, "Load KML File");
	dialog.setFilter(tr("All|*.kml *.mrd *.log"));

	config.setGroup("Data");
	if (!config.hasKey("dir"))
		config.writeEntry("dir", "/mnt/card");

	dialog.setDirectory(config.readEntry("dir"));
	dialog.setDirectorySelectMode(false);
	dialog.setHomeDirectoryCdUpEnable(true);

	if (dialog.exec() == QDialog::Accepted) {
		QFile routeFile;
		bool loadKml = false;
		if (dialog.getFileName().right(4) == ".log") {
			_gpsLogHandler->loadObsolateLogFile(dialog.getFileName());
			paintMaps(_centerNS, _centerEW);
			return;
		}

		if (dialog.getFileName().right(4) == ".kml") {
			loadKml = true;
		}
		routeFile.setName(dialog.getFileName());
		if (routeFile.exists()) {
			if (!routeFile.open(IO_ReadOnly)) {
				QMessageBox::warning(this, "Load error",
						"Could not load file.\n" + dialog.getFileName());
				return;
			}
		} else {
			QMessageBox::warning(this, "Load error", "File not found.\n"
					+ dialog.getFileName());
			return;
		}

		char buf[512];
		if (loadKml) {
			while (!routeFile.atEnd()) {
				int sz = routeFile.readLine(buf, sizeof(buf));
				if (sz <= 0)
					break;
				if (strstr(buf, "<coordinates>") != NULL)
					break;
			}
		}

		_route.clear();
		Radian radNS, radEW;
		while (!routeFile.atEnd()) {
			int sz = routeFile.readLine(buf, sizeof(buf));
			if (sz <= 0)
				break;
			const QStringList& qslTemp = QStringList::split(",", buf);
			int intHight = 0;
			if (loadKml && qslTemp.count() >= 3) {
				intHight = qslTemp[2].toInt();
			}
			if (qslTemp.count() >= 2) {
				radEW.parse2("E" + qslTemp[0]);
				radNS.parse2("N" + qslTemp[1]);
				Coordinate coordinate(radEW, radNS, intHight);
				if (loadKml)
					coordinate.convWGS84ToTokyo();
				_route.push_back(coordinate);
			}
		}
		routeFile.close();

		// ̺
		paintMaps(_centerNS, _centerEW);
	}
}

void Zega::resizeEvent(QResizeEvent *event) {
	if (_mapBuffer->width() != this->width() || _mapBuffer->height()
			!= this->height()) {
		_mapBuffer->resize(this->width(), this->height());
		paintMaps(_centerNS, _centerEW);
	}
}

void Zega::mousePressEvent(QMouseEvent *event) {
#ifdef _DEBUG_PRINT
	cout << "mousePressEvent " << event->pos().x() << ", " << event->pos().y() << endl;
#endif

	Radian radNS, radEW;

	radNS.setData(_centerNS.toLong() + ((_pixelNS * (event->pos().y() - _mapBuffer->height()/2))>>_pixelBase));
	radEW.setData(_centerEW.toLong() + ((_pixelEW * (event->pos().x() - _mapBuffer->width()/2))>>_pixelBase));

	paintMaps(radNS, radEW);
}

void Zega::keyPressEvent(QKeyEvent *event) {
	Radian radNS, radEW;
	int intWait = 2;

	radNS = _centerNS;
	radEW = _centerEW;

	if (event->state() & ShiftButton) {
		intWait = 1;
	}

	switch (event->key()) {
	case Key_Up:
		radNS.setData(_centerNS.toLong() - (((_pixelNS * _mapBuffer->height())
				/intWait) >> _pixelBase));
		paintMaps(radNS, radEW);
		break;
	case Key_Down:
		radNS.setData(_centerNS.toLong() + (((_pixelNS * _mapBuffer->height())
				/intWait) >> _pixelBase));
		paintMaps(radNS, radEW);
		break;
	case Key_Left:
		radEW.setData(_centerEW.toLong() - (((_pixelEW * _mapBuffer->width())
				/intWait) >> _pixelBase));
		paintMaps(radNS, radEW);
		break;
	case Key_Right:
		radEW.setData(_centerEW.toLong() + (((_pixelEW * _mapBuffer->width())
				/intWait) >> _pixelBase));
		paintMaps(radNS, radEW);
		break;
	case Key_Space:
		if (_gpsLogHandler->isAlive()) {
			Coordinate coordinate = _gpsLogHandler->getNowCoordinate();
			radNS = coordinate.getRadNS();
			radEW = coordinate.getRadEW();
			paintMaps(radNS, radEW);
		} else {
			return;
		}
		break;
	case Key_1:
		if (mapDirs.count() > 0)
			slotChScale(0);
		break;
	case Key_2:
		if (mapDirs.count() > 1)
			slotChScale(1);
		break;
	case Key_3:
		if (mapDirs.count() > 2)
			slotChScale(2);
		break;
	case Key_4:
		if (mapDirs.count() > 3)
			slotChScale(3);
		break;
	case Key_5:
		if (mapDirs.count() > 4)
			slotChScale(4);
		break;
	case Key_6:
		if (mapDirs.count() > 5)
			slotChScale(5);
		break;
	case Key_7:
		if (mapDirs.count() > 6)
			slotChScale(6);
		break;
	case Key_8:
		if (mapDirs.count() > 7)
			slotChScale(7);
		break;
	case Key_9:
		if (mapDirs.count() > 8)
			slotChScale(8);
		break;
	case Key_0:
		if (mapDirs.count() > 9)
			slotChScale(9);
		break;
	case Key_G:
		if (_gpsLogHandler->isStopped())
			startGps();
		else
			stopGps();
		break;
	case Key_C:
		slotToggleAutoCentered();
		break;
	case Key_H:
		slotToggleShowGpsLog();
		break;
	case Key_R:
		slotToggleShowRoute();
		break;
	case Key_L:
		slotLoad();
		break;
	case Key_D:
		_gpsLogHandler->changeDebugPrint();
		break;
	default:
		break;
	}
}
void Zega::slotToggleShowGpsLog() {
	if (_showGpsLog)
		_showGpsLog = false;
	else
		_showGpsLog = true;
	paintMaps(_centerNS, _centerEW);
}

void Zega::slotToggleShowRoute() {
	if (_showRoute)
		_showRoute = false;
	else
		_showRoute = true;
	paintMaps(_centerNS, _centerEW);
}

void Zega::slotToggleAutoCentered() {
	if (_autoCentered)
		_autoCentered = false;
	else
		_autoCentered = true;
}

void Zega::paintEvent(QPaintEvent*) {
	if (_mapBuffer->width() == 0) {
		_mapBuffer->resize(this->width(), this->height());
		paintMaps(_centerNS, _centerEW);
	} else {
		bitBlt(this, 0, 0, _mapBuffer);
	}
}

// ץɽ
void Zega::about() {
	QMessageBox::about(
			this,
			"About this program",
			tr("Zega Version 0.6.1\nThis is a GPS assistance program for Zaurus.\nCopyright HB1 2008.06.06"));
}

void Zega::paintMaps(Radian radNS, Radian radEW) {
	Radian radMapNS, radMapEW;
	Radian radMapWidth, radMapHeight;
	Radian radMinNS, radMinEW;
	QString mapFileName;
	QString tempStr;

	// 濴ηٰ٤Ͽ
	_centerNS = radNS;
	_centerEW = radEW;

	// ;
	_bsl = _mapBuffer->width();
	_bsu = _mapBuffer->height();
	_bsr = 0;
	_bsd = 0;

	// Ͽ޽񤭹ߥХåեꥢ
	_mapBuffer->fill();

	mapFileName = _mapList->searchMapFileName(radNS.toLong(), radEW.toLong(),
			&radMapEW, &radMapNS, &radMapWidth, &radMapHeight);

#ifdef _DEBUG_PRINT
	cout << "mapFileName: " << mapFileName << endl;
	cout << "mapNS: " << radMapNS.toNS() << endl;
	cout << "mapEW: " << radMapEW.toEW() << endl;
	cout << "mapMapHeight: " << radMapHeight.toString() << endl;
	cout << "mapMapWidth: " << radMapWidth.toString() << endl;
#endif

	if (!mapFileName.isEmpty()) {
		radMinEW.setData(radMapEW.toLong() - radMapWidth.toLong()/2);
		radMinNS.setData(radMapNS.toLong() - radMapHeight.toLong()/2);

		_buffer = _mapCache->getMap(mapFileName);
		if (_buffer != NULL) {
			int x = (radEW.toLong() - radMinEW.toLong())* 1.0
					*(_buffer->width()) /radMapWidth.toLong();
			int y = (radNS.toLong() - radMinNS.toLong())* 1.0
					*(_buffer->height()) /radMapHeight.toLong();
			int isx = x - _mapBuffer->width()/2;
			_bsl = 0;
			if (isx < 0) {
				_bsl = _mapBuffer->width()/2 - x;
				isx = 0;
			}
			int isy = y - _mapBuffer->height()/2;
			_bsu = 0;
			if (isy < 0) {
				_bsu = _mapBuffer->height()/2 - y;
				isy = 0;
			}

			// Ͽ޲ԡ
			bitBlt(_mapBuffer, _bsl, _bsu, _buffer, isx, isy, _buffer->width()
					- isx, _buffer->height() - isy);

			// ;׻
			_bsr = _bsl + _buffer->width() - isx;
			_bsd = _bsu + _buffer->height() - isy;
		}

#ifdef _DEBUG_PRINT
		cout << "compute pixel" << endl;
#endif
		// 1 pixel ٰ٤Ͽ
		int bh, bw;
		if (_buffer != NULL) {
			bh = _buffer->height();
			bw = _buffer->width();
		} else {
			bh = 500;
			bw = 500;
		}
		if (radMapWidth.toLong() < (+2147483647L >> PIXEL_BASE)
				&& radMapHeight.toLong() < (+2147483647L >> PIXEL_BASE)) {
			_pixelBase = PIXEL_BASE;
		} else {
			_pixelBase = 0;
		}

		_pixelNS = (radMapHeight.toLong() << _pixelBase) / bh;
		_pixelEW = (radMapWidth.toLong() << _pixelBase) / bw;
#ifdef _DEBUG_PRINT
		cout << "pixelNS: " << _pixelNS << endl;
		cout << "pixelEW: " << _pixelEW << endl;
#endif
	}

#ifdef _DEBUG_PRINT
	cout << "copy other maps" << endl;
#endif
	copyMap( 1, 0, radEW, radNS);
	copyMap(-1, 0, radEW, radNS);
	copyMap( 0, 1, radEW, radNS);
	copyMap( 0, -1, radEW, radNS);
	copyMap( 1, 1, radEW, radNS);
	copyMap( 1, -1, radEW, radNS);
	copyMap(-1, 1, radEW, radNS);
	copyMap(-1, -1, radEW, radNS);

#ifdef _DEBUG_PRINT
	cout << "compute paint range" << endl;
#endif
	// ϰϤ١٤Ѵ
	Coordinate coordinate;
	long cEW = _centerEW.toLong();
	long cNS = _centerNS.toLong();
	long x, y, x2, y2;
	int cx = this->width()/2;
	int cy = this->height()/2;
	long lEW = _centerEW.toLong() - ((_pixelEW * cx) >> _pixelBase);
	long uEW = _centerEW.toLong() + ((_pixelEW * cx) >> _pixelBase);
	long lNS = _centerNS.toLong() - ((_pixelNS * cy) >> _pixelBase);
	long uNS = _centerNS.toLong() + ((_pixelNS * cy) >> _pixelBase);
	QPainter painter(_mapBuffer);

	if (_showRoute) {
#ifdef _DEBUG_PRINT
		cout << "draw route" << endl;
#endif
		// 롼Ȥ
		long lastEW, lastNS, nowEW, nowNS;
		long aEW, aNS, bEW, bNS;
		bool writeRoute = false;
		bool drawNewLine;
		painter.setPen(QPen(_routeColor, _routeWidth, Qt::SolidLine));
		vector<Coordinate>::iterator it = _route.begin();
		if (it != _route.end()) {
			lastEW = (*it).getRadEW().toLong();
			lastNS = (*it).getRadNS().toLong();
			it++;
			if (lastEW >= lEW && lastEW <= uEW && lastNS >= lNS && lastNS
					<= uNS) {
				writeRoute = true;
			}
		}
		while (it != _route.end()) {
			drawNewLine = false;
			nowEW = (*it).getRadEW().toLong();
			nowNS = (*it).getRadNS().toLong();
			if (nowEW >= lEW && nowEW <= uEW && nowNS >= lNS && nowNS <= uNS) {
				drawNewLine = true;
				writeRoute = true;
			} else if (writeRoute) {
				drawNewLine = true;
				writeRoute = false;
			} else {
				writeRoute = false;

				if (lastNS > nowNS) {
					aNS = nowNS;
					aEW = nowEW;
					bNS = lastNS;
					bEW = lastEW;
				} else {
					aNS = lastNS;
					aEW = lastEW;
					bNS = nowNS;
					bEW = nowEW;
				}
				if (aNS < lNS && bNS > lNS) {
					if (aEW < lEW) {
						if (bEW > lEW)
						drawNewLine = true;
					} else if (aEW < uEW) {
						drawNewLine = true;
					} else {
						if (bEW < uEW)
						drawNewLine = true;
					}
				}
				if (aNS < uNS && bNS > uNS) {
					if (bEW < lEW) {
						if (aEW > lEW)
						drawNewLine = true;
					} else if (bEW < uEW) {
						drawNewLine = true;
					} else {
						if (aEW < uEW)
						drawNewLine = true;
					}
				}

				if (drawNewLine == false) {
					if (lastEW > nowEW) {
						aNS = nowNS;
						aEW = nowEW;
						bNS = lastNS;
						bEW = lastEW;
					} else {
						aNS = lastNS;
						aEW = lastEW;
						bNS = nowNS;
						bEW = nowEW;
					}
					if (aEW < lEW && bEW > lEW) {
						if (aNS < lNS) {
							if (bNS > lNS)
							drawNewLine = true;
						} else if (aNS < uNS) {
							drawNewLine = true;
						} else {
							if (bNS < uNS)
							drawNewLine = true;
						}
					}
					if (aEW < uEW && bEW > uEW) {
						if (bNS < lNS) {
							if (aNS > lNS)
							drawNewLine = true;
						} else if (bNS < uNS) {
							drawNewLine = true;
						} else {
							if (aNS < uNS)
							drawNewLine = true;
						}
					}
				}
			}
			if (drawNewLine) {
				x = ((lastEW - cEW) << _pixelBase) / _pixelEW + cx;
				y = ((lastNS - cNS) << _pixelBase) / _pixelNS + cy;
				x2 = ((nowEW - cEW) << _pixelBase) / _pixelEW + cx;
				y2 = ((nowNS - cNS) << _pixelBase) / _pixelNS + cy;
				painter.drawLine(x, y, x2, y2);
			}
			lastEW = nowEW;
			lastNS = nowNS;
			it++;
		}
	}

	if(_showGpsLog) {
#ifdef _DEBUG_PRINT
		cout << "point obsolute logs" << endl;
#endif
		// ľʳΥݥ
		painter.setPen(QPen(blue, 1, SolidLine));
		painter.setBrush(QBrush(_obsoluteLogColor, SolidPattern));
		_gpsLogHandler->setObsolateLogFirst();
		while (!_gpsLogHandler->endObsolateLog()) {
			coordinate = _gpsLogHandler->getNextObsolateLog();

			// ֤Ͽޤ˥ݥ
			if (coordinate.getRadEW().toLong() >= lEW && coordinate.getRadEW().toLong() <= uEW && coordinate.getRadNS().toLong() >= lNS && coordinate.getRadNS().toLong() <= uNS) {
				x = ((coordinate.getRadEW().toLong() - cEW) << _pixelBase) / _pixelEW + cx;
				y = ((coordinate.getRadNS().toLong() - cNS) << _pixelBase) / _pixelNS + cy;
				painter.drawEllipse(x - _logWidth/2, y - _logWidth/2, _logWidth, _logWidth);
			}
		}

#ifdef _DEBUG_PRINT
		cout << "point recent logs" << endl;
#endif
		// ľΥݥ
		painter.setBrush(QBrush(_recentLogColor, SolidPattern));
		_gpsLogHandler->setRecentLogFirst();
		while (!_gpsLogHandler->endRecentLog()) {
			coordinate = _gpsLogHandler->getNextRecentLog();

			// ֤Ͽޤ˥ݥ
			if (coordinate.getRadEW().toLong() >= lEW && coordinate.getRadEW().toLong() <= uEW && coordinate.getRadNS().toLong() >= lNS && coordinate.getRadNS().toLong() <= uNS) {
				x = ((coordinate.getRadEW().toLong() - cEW) << _pixelBase) / _pixelEW + cx;
				y = ((coordinate.getRadNS().toLong() - cNS) << _pixelBase) / _pixelNS + cy;
				painter.drawEllipse(x - _logWidth/2, y - _logWidth/2, _logWidth, _logWidth);
			}
		}
	}

#ifdef _DEBUG_PRINT
	cout << "draw center cross" << endl;
#endif
	painter.setPen(QPen(black, 1, SolidLine));
	painter.drawLine(_mapBuffer->width()/2 - 10, _mapBuffer->height()/2,
	_mapBuffer->width()/2 + 10, _mapBuffer->height()/2);
	painter.drawLine(_mapBuffer->width()/2, _mapBuffer->height()/2 - 10,
	_mapBuffer->width()/2, _mapBuffer->height()/2 + 10);
	bitBlt(this, 0, 0, _mapBuffer);
}

// Ͽɹ
bool Zega::copyMap(int x, int y, Radian radEW, Radian radNS) {
	Radian radMapEW, radMapNS;
	Radian radMinEW, radMinNS;
	Radian radMapWidth, radMapHeight;
	QString mapFileName;

	if (x < 0 && _bsl <= 0)
		return false;
	if (x > 0 && _bsr >= _mapBuffer->width())
		return false;
	if (y < 0 && _bsu <= 0)
		return false;
	if (y > 0 && _bsd >= _mapBuffer->height())
		return false;

	mapFileName = _mapList->pointMapFileName(y, x, &radMapEW, &radMapNS,
			&radMapWidth, &radMapHeight);
	if (!mapFileName.isEmpty()) {
		radMinEW.setData(radMapEW.toLong() - radMapWidth.toLong()/2);
		radMinNS.setData(radMapNS.toLong() - radMapHeight.toLong()/2);

		_buffer = _mapCache->getMap(mapFileName);
		if (_buffer == NULL) {
			return false;
		}

		int w = (radEW.toLong() - radMinEW.toLong())*(_buffer->width())
				/radMapWidth.toLong();
		int h = (radNS.toLong() - radMinNS.toLong())*(_buffer->height())
				/radMapHeight.toLong();
		int isx = w - _mapBuffer->width()/2;
		int bsx = 0;
		if (isx < 0) {
			bsx = _mapBuffer->width()/2 - w;
			isx = 0;
		}
		int isy = h - _mapBuffer->height()/2;
		int bsy = 0;
		if (isy < 0) {
			bsy = _mapBuffer->height()/2 - h;
			isy = 0;
		}

		int copy_width = _buffer->width() - isx;
		if (x > 0 && _bsr > bsx) {
			isx += _bsr - bsx;
			copy_width -= _bsr - bsx;
			bsx = _bsr;
		} else if (x < 0 && copy_width > _bsl - bsx)
{
			copy_width = _bsl - bsx;
		}

					int copy_height = _buffer->height() - isy;
		if (y > 0 && _bsd > bsy) {
			isy += _bsd - bsy;
			copy_height -= _bsd - bsy;
			bsy = _bsd;
		} else if (y < 0 && copy_height > _bsu - bsy)
{
			copy_height = _bsu - bsy;
		}

		// Ͽ޲ԡ
					bitBlt(_mapBuffer, bsx, bsy, _buffer, isx, isy, copy_width,
					copy_height);
	}
	return true;
}
